# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

use strict;
use warnings;
use AI::MXNet qw(mx);
use Test::More tests => 2;

sub test_bulk
{
    my $x;
    mx->engine->bulk(10, sub {
        $x = mx->nd->ones([10]);
        $x *= 2;
        $x += 1;
        $x->wait_to_read();
        $x += 1;
        ok(($x->aspdl == 4)->all);
        for my $i (1..100)
        {
            $x += 1;
        }
    });
    ok(($x->aspdl == 104)->all);
}

test_bulk();