# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

package AI::MXNet::AutoLoad;
use strict;
use warnings;

sub AUTOLOAD
{
    my ($class) = @_;
    my ($prefix, $real_class) = $class->config;
    my ($name) = our $AUTOLOAD =~ /::(\w+)$/;
    my $sub = "_${prefix}_$name";
    {
        no strict 'refs';
        *{"${class}::$name"} = sub { shift; $real_class->$sub(@_); };
    }
    goto $class->can($name);
}

1;
